package geometry;



/**
 * A point in double-precision doubleing-point 3 space.
 * The point is internally represented as a point in homogeneous
 * coordinate four-space, so that the range of transformations that
 * can be applied to it include perspective projection. However,
 * in nearly all cases the user need not be aware of the internal
 * representation, and can regard it as a 3-tuple (x,y,z).
 * @author Richard Lobb
 * @version 1.1, April 2001
 */


public class Point3d {

/**
 * The internal representation of the point (as a
 * vector in homogeneous coordinate 4-space).
 */
	protected double[] coord = new double[4];

    // Constructors
    // ============

/**
 * Constructs a point at the origin (0,0,0)
 */
    public Point3d() {
		coord[0] = coord[1] = coord[2] = 0;
		coord[3] = 1;
	}

/**
 * Constructs the point (x,y,z)
 * @param x The x-coordinate of the point
 * @param y The y-coordinate of the point
 * @param z The z-coordinate of the point
 */
	public Point3d(double x, double y, double z) {
		coord[0] = x; coord[1] = y; coord[2] = z; coord[3] = 1;
	}

/**
 * Constructs a point from a 3-vector or 4-vector of doubles.
 * @param p The double array containing the three or four coordinates
 * of the point. The created point is <code>(p[0],p[1],p[2],1)</code>
 * or <code>(p[0],p[1],p[2],p[3])</code>
 * respectively.
 * @throws RuntimeException if the array length is neither 3 nor 4.
 */
    public Point3d(double[] p) {
        if (p.length != 3 && p.length != 4)
            throw new RuntimeException("Point3d: Constructor needs 3 or 4 doubles");
        else {
            coord[0] = p[0]; coord[1] = p[1]; coord[2] = p[2];
            if (p.length == 3)
                coord[3] = 1;
            else
                coord[3] = p[3];
        }
    }

/**
 * Constructs a copy of an existing point.
 * @param p The point to be "cloned".
 */
	public Point3d(Point3d p) {
		for (int i = 0; i < 4; i++)
			coord[i] = p.coord[i];
	}

/**
 * Constructs a point given its position vector.
 * @param v The position vector of the point to be constructed.
 * The created point is (v.dx, v.dy, v.dz).
 */
	public Point3d(Vector3d v) {
    // Constructor, given the point's position vector
		coord[0] = v.dx; coord[1] = v.dy;
		coord[2] = v.dz; coord[3] = 1;
	}

/**
 * Constructs a point as a convex sum of two other points.
 * The constructed point is: (1 - alpha) p1 + alpha p2
 * @param p1 The first point.
 * @param p2 The second point
 * @param alpha The "mixing coefficient"
 */
	public Point3d(Point3d p1, Point3d p2, double alpha) {
		double beta = 1 - alpha;
        coord[0] = beta * p1.x() + alpha * p2.x();
		coord[1] = beta * p1.y() + alpha * p2.y();
		coord[2] = beta * p1.z() + alpha * p2.z();
        coord[3] = 1;
	}

    // Component access methods
    // ========================

/**
 * The <i>x</i> coordinate of the point
 */
	public double x() { return coord[0] / coord[3]; }
/**
 * The <i>y</i> coordinate of the point
 */
	public double y() { return coord[1] / coord[3]; }
/**
 * The <i>z</i> coordinate of the point
 */
	public double z() { return coord[2] / coord[3]; }

/**
 * Copies the coordinates of "this" into the given double array.
 * @param p the target array. This must be an array of
 * either 3 or 4 doubles. If p is a 3-element array, its values are set to
 * the (projected) values <i>x</i>, <i>y</i> and <i>z</i>.
 * If p is a 4-element array, all 4 internal
 * values (homogeneous coordinates) are copied.
 * @throws RuntimeException if the array length is neither 3 nor 4.
*/
    public void copyInto(double[] p) {
        if (p.length == 3) {
            p[0] = coord[0] / coord[3];
            p[1] = coord[1] / coord[3];
            p[2] = coord[2] / coord[3];
        }
        else if (p.length == 4) {
            for (int i = 0; i < 4; i++)
                p[i] = coord[i];
        }
        else
            throw new RuntimeException("Point3d.copyInto: needs 3 or 4-element target");
    }

 /**
 * The position vector of the point
 * @return a new Vector3d object (this - (0,0,0))
 */
   public Vector3d positionVector() {
        return new Vector3d(x(),y(),z());
    }

 /**
 * A 3-element float array representing "this"
 * @return a new double array {x, y, z}. This form
 * is useful for passing to the OpenGL engine.
 */
   public float[] get3fv() {
    // Return point as a 3-element array of doubles
        return new float[] {(float) x(), (float) y(), (float) z()};
    }

 /**
  * A string represention of "this".
  * @return a string represention of "this", such as "(2.001, 3, 4)".
  * The individual components are of whatever length the standard Java
  * double-to-string conversion routines deem appropriate
  */
    public String toString() {
    // convert point to a string
        return "(" + x() + "," + y() + "," + z() + ")";
    }

    // Functional style operations
    // ===========================

 /**
 * A new point equal to "this" plus a displacement vector.
 * @param v the vector to be added to "this"
 * @return a new Vector3d object this + v
 */
	public Point3d plus(Vector3d v) {
    // point + vector -> point
		return new Point3d(x()+v.dx(), y()+v.dy(), z()+v.dz());
	}

 /**
 * A new point equal to "this" minus a displacement vector.
 * @param v the vector to be subtracted from "this"
 * @return a new Vector3d object equal to "this" - v
 */
	public Point3d minus(Vector3d v) {
	// point - vector -> point
		return new Point3d(x()-v.dx(), y()-v.dy(), z()-v.dz());
	}

 /**
  * A new vector representing the displacement from p to "this"
  * @return a new Vector3d object equal to "this" - p
  */
    public Vector3d minus(Point3d p) {
    // Returns the vector this - p
        return new Vector3d(x() - p.x(), y() - p.y(), z() - p.z());
    }

 /**
  * A new point a specified relative distance from "this" to another point.
  * @param alpha the relative distance from this to p. If alpha
  * is in the range [0,1], the returned point lies on the line segment
  * connecting "this" and p. Otherwise, it lies on the straight line
  * through the two points, but outside the segment between them.
  * @param p the other point
  * @return a new point (1-alpha) * this + alpha * p
  */
    public Point3d interpolation(double alpha, Point3d p) {
        return new Point3d(this, p, alpha);
    }

 /**
  * The result of transforming "this" by the given transformation
  * @param trans the transformation to be applied
  */
   public Point3d transformedBy(Transformation trans) {
        double[][] m = trans.matrixd();
        Point3d result = new Point3d();
        for (int i = 0; i < 4; i++) {
            double sum = 0;
            for (int j = 0; j < 4; j++)
                sum += coord[j] * trans.m[i][j];
            result.coord[i] = sum;
        }
        return result;
    }

/**
  * True if and only if the x, y and z components of "this" are equal to the corresponding
  * components of another point to within a specified tolerance.
  * Note that this comparison
  * is performed in the projected 3-space, not the homogeneous coordinate 4-space.
  * @param p The other point
  * @param tolerance The maximum permitted absolute difference between any coordinate of "this"
  * and the corresponding coordinate of "p" for the result returned to be true.
  * @return True if and only if this.x == p.x, this.y == p.y and
  * this.z == p.z, where all equality tests are to within the specified tolerance.
  */
	public boolean epsilonEquals(Point3d p, double tolerance) {
		return
          Math.abs(x() - p.x()) <= tolerance &&
          Math.abs(y() - p.y()) <= tolerance &&
          Math.abs(z() - p.z()) <= tolerance;
	}

    // Non-functional (mutable) operations
    // ===================================

/**
 * Sets the coordinates of "this" to be equal to those of a given point
 * @param p the point whose coordinates are to be copied
 */
   public void set(Point3d p) {
        // Sets this to have the same value as p.
        coord[0] = p.coord[0];
        coord[1] = p.coord[1];
        coord[2] = p.coord[2];
        coord[3] = p.coord[3];
    }

/**
 * Sets the coordinates of "this" to given values. The internal
 * fourth coordinate ("w") is set to 1.
 * @param x the x coordinate
 * @param y the y coordinate
 * @param z the z coordinate
 */
	public void set(double x, double y, double z) {
		coord[0] = x;
		coord[1] = y;
		coord[2] = z;
		coord[3] = 1;
	}

/**
 * Sets the coordinates of "this" to the given homogeneous coordinates.
 * @param x the x coordinate
 * @param y the y coordinate
 * @param z the z coordinate
 * @param w the w coordinate
 */
	public void set(double x, double y, double z, double w) {
		// Sets this to (x,y,z,w)
		coord[0] = x;
		coord[1] = y;
		coord[2] = z;
		coord[3] = w;
	}

	private void normalise() {
	// Projects onto the plane w = 1 (projector through origin).
		if (coord[3] != 1) {
			coord[0] /= coord[3];
			coord[1] /= coord[3];
			coord[2] /= coord[3];
			coord[3] = 1;
		}
    }

 /**
  * Transforms "this" by a given transformation
  * @param trans the transformation to be applied
  */
   public void transform(Transformation trans) {
        double[][] m = trans.matrixd();
        double xnew, ynew, znew, wnew;
        xnew = m[0][0] * coord[0] + m[0][1] * coord[1] + m[0][2] * coord[2] + m[0][3] * coord[3];
        ynew = m[1][0] * coord[0] + m[1][1] * coord[1] + m[1][2] * coord[2] + m[1][3] * coord[3];
        znew = m[2][0] * coord[0] + m[2][1] * coord[1] + m[2][2] * coord[2] + m[2][3] * coord[3];
        wnew = m[3][0] * coord[0] + m[3][1] * coord[1] + m[3][2] * coord[2] + m[3][3] * coord[3];
        set(xnew, ynew, znew, wnew);
    }
}
