package geometry;

import java.awt.*;
import java.awt.event.*;

/**
 * A trackball to provide camera control for use in terrain display.
 * The trackball is invisibly associated with an AWT component,
 * which is specified in the constructor.
 * The trackball maintains the position and orientation of a
 * camera platform. Left-button drags over the controlled 
 * component adjust the orientation of the camera platform.
 * The effect depends on which of left-right or up-down mouse motion
 * predominates. Predominant up-down mouse motion causes the lookat
 * direction to change in elevation only. Predominant left-right motion
 * changes lookat azimuth angle only.
 *
 * Right button drags over the controlled
 * component move the viewing platform. The up-down component
 * of the mouse motion results in motion in the "look at"
 * direction, while the left-right component results in norizontal
 * motion perpendicular to the look at direction.
 *  
 * @author Richard Lobb
 * @version 1.0, Dec 2001
 */

public class TerrainTrackball {
    private static final Vector3f VERTICAL = new Vector3f(0,1,0);
    private Component component; // Thing we're "attached" to
    private Point3f cameraPos;  // Where the camera is
    private Vector3f cameraZInitial;    // Where the camera is pointing initially (LH coordinate system!!!)
    private Vector3f cameraXInitial;    // The initial camera x direction
    private Vector3f cameraYInitial;    // The initial camera y direction
    Transformation totalRotation = new Transformation();    // The rotation to apply to the camera platform
    float panScaleFactor;
    private Vector3f v0, v1;    // Vectors from centre of trackball
    private boolean isRightButton = false;
        
/**
 * Constructs a trackball that operates (invisibly) over the screen area
 * occupied by the specified component.
 * @param component The AWT component that defines the area over which
 * the trackball operates.
 * @param initialCameraPos The initial camera position
 * @param initialLookatDirection The direction in which the camera is looking initially
 * @param panScaleFactor The amount by which the camera moves when the mouse
 * is dragged over the full width of the component.
 */
    public TerrainTrackball(Component component,
                     Point3f initialCameraPos,
                     Vector3f initialLookatDirection,
                     float panScaleFactor) {

        this.component = component;
        cameraPos = initialCameraPos;
        cameraZInitial = initialLookatDirection.normalised();
        cameraXInitial = cameraZInitial.cross(new Vector3f(0,1,0)).normalised();
        cameraYInitial = cameraXInitial.cross(cameraZInitial);
        this.panScaleFactor = panScaleFactor;
        component.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent e) {
                // Save the vector from the trackball centre to the start point of the
                // drag (i.e. where the mouse is clicked) in instance variable v0
                v0 = makeTrackballVector(e);
                isRightButton = (e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0;
            }
        });
        component.addMouseMotionListener(new MouseMotionAdapter() {
            public void mouseDragged(MouseEvent e) {
                handleDrag(e);
            }
        });
    }

/** Returns the current camera position in OpenGL format. */     
    public float[] getGLCameraPos() {
        return cameraPos.get3fv();
    }
    
/** Returns the current camera position as a Point3f. */     
    public Point3f getCameraPos() {
        return cameraPos;
    }
 
/** Returns the lookat point in OpenGL format.
    The lookat point is defined to be the camera position
    plus a unit vector in the lookat direction. */     
    public float[] getGLLookatPoint() {
        return cameraPos.plus(totalRotation.times(cameraZInitial)).get3fv();
    }
    
/** Returns the lookat point.
    The lookat point is defined to be the camera position
    plus a unit vector in the lookat direction. */     
    public Point3f getLookatPoint() {
        return cameraPos.plus(totalRotation.times(cameraZInitial));
    }

/** Returns the lookat vector in OpenGL format, i.e. the camera z in a LH coord sys */
    public float[] getGLLookat() {
        return totalRotation.times(cameraZInitial).get3fv();
    }

/** Returns the camera x axis vector in OpenGL format, i.e. the -ve of camera z */
    public float[] getGLRight() {
        return totalRotation.times(cameraXInitial).get3fv();
    }
/** Returns the up vector in OpenGL format.
    The lookat point is defined to be the camera position
    plus a unit vector in the lookat direction. */     
    public float[] getGLUp() {
        return totalRotation.times(cameraYInitial).get3fv();
    }
    
/** Returns the up vector. */     
    public Vector3f getUp() {
        return totalRotation.times(cameraYInitial);
    }

/** Set the camera position to the given value */
    public void setPosition(Point3f pos) {
        cameraPos = pos;
    }

/** Multiply the panScaleFactor (i.e. "speed") by a given factor */
    public void multiplySpeedBy(float factor) {
        panScaleFactor *= factor;
    }

    private void handleDrag(MouseEvent e) {
        // Handle mouse drags. Called directly by MouseMotionAdaptor.
        Vector3f v1 = makeTrackballVector(e);
        float dx = v1.dx() - v0.dx();
        float dy = v1.dy() - v0.dy();
        if (isRightButton)  {   // Right button - move camera platform
            Vector3f forwardMotion = cameraZInitial.times(dy);
            Vector3f sidewaysMotion = cameraXInitial.times(dx);
            Vector3f totalMotion = forwardMotion.plus(sidewaysMotion).times(panScaleFactor/2);
            cameraPos = cameraPos.plus(totalRotation.times(totalMotion));
        }
        else  { // Left button -- rotate camera view direction
            if (Math.abs(dx) > Math.abs(dy)) {
                totalRotation.multiplyOnLeft(Transformation.rotation(-90 * dx, VERTICAL));
            } else {
                Vector3f currentCameraX = totalRotation.times(cameraXInitial);
                totalRotation.multiplyOnLeft(Transformation.rotation(90 * dy, currentCameraX));
            }
        }
        v0 = v1;
    }

    private Vector3f makeTrackballVector(MouseEvent e) {
        Point p = e.getPoint();
        Dimension dim = e.getComponent().getSize();
        int width = dim.width;
        int height = dim.height;

        float x = ( 2 * p.x - width ) / (float) width;
        float y = ( height - 2 * p.y ) / (float) height;
        float tmp = 1 - x*x - y*y;
        float z = tmp <= 0 ? 0 : (float) Math.sqrt(tmp);
        return new Vector3f(x, y, z).normalised();
    }
}
