package geometry;

import java.awt.*;
import java.awt.event.*;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import TTPproject.CONSTANTS;
import TTPproject.Configuration;

/**
 * A trackball to provide rotations for use in 3D scene display.
 * The trackball is invisibly associated with an AWT component,
 * which is specified in the constructor. When the mouse is dragged
 * over that component, the internal rotation matrix is adjusted
 * accordingly and the Component's "repaint" method is called.
 * The component is expected to request the trackball for its
 * current rotation. The rotation is made available either as a
 * <i>Transformation</i> or as an OpenGL-compatible float array (matrix).
 * The trackball also supports a "zoom" capability -- dragging up
 * and down the window with the right mouse button down respectively
 * increases/decreases an internally-maintained zoom factor.
 * @author Richard Lobb
 * @version 1.1, May 2001
 */

public class Trackball 
    implements CONSTANTS, MouseWheelListener{

/**
 * The default maximum zoom factor
 */
    protected static final float MAX_ZOOM_DEFAULT = 50;
/**
 * The default minimum zoom factor
 */
    protected static final float MIN_ZOOM_DEFAULT = 0.02f;

    private static final float ZOOM_RATE = 1;
    
    
    private Component component; // Thing to repaint after each rotation  ...
    private boolean repaintReqd = true; // ... unless this is false.
    private Vector3f v0, v1;    // Vectors from centre of trackball
    private boolean modifiers_keydown= false;
    private boolean first_time_moved= true;
    private Transformation totalRotation = new Transformation(); // Identity
    private float zoomFactor;
    
    //new zoom in 
    private float lengthFactor;

    private float maxZoom = MAX_ZOOM_DEFAULT;
    private float minZoom = MIN_ZOOM_DEFAULT;
    private boolean isRightButton = false;

    private Configuration config;

/**
 * Constructs a trackball that operates (invisibly) over the screen area
 * occupied by the specified component.
 * @param component The AWT component that defines the area over which
 * the trackball operates. Changes to the rotation matrix or zoom factor
 * result in a call to the component's "repaint" method.
 */
    public Trackball(Component component, Configuration config) {
        this(component, MIN_ZOOM_DEFAULT, MAX_ZOOM_DEFAULT, config);
	
    }

/**
 * Constructs a trackball that operates (invisibly) over the screen area
 * occupied by the specified component, and with a specified allowable
 * range of zoom factors.
 * @param component The AWT component that defines the area over which
 * the trackball operates. Changes to the rotation matrix or zoom factor
 * result in a call to the component's "repaint" method.
 * @param minZoom The minimum allowable zoom factor. Mouse drags that
 * attempt to reduce the zoom factor below this value are ignored.
 * @param maxZoom The maximum allowable zoom factor. Mouse drags that
 * attempt to increase the zoom factor above this value are ignored.
 */
    public Trackball(Component component, float minZoom, float maxZoom, Configuration config) {
        this.component = component;

	//initialize 2 vectors
	//v0= new Vector3f(); 
	//v1= new Vector3f();

        component.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent e) {
                // Save the vector from the trackball centre to the start point of the
                // drag (i.e. where the mouse is clicked) in instance variable v0
		if(TTPproject.Ttp.commentslevel >= 3){
		    System.out.println("Trackball.java: add mousePressed");
		}
                v0 = makeTrackballVector(e);
		
		if(TTPproject.Ttp.commentslevel >= 3){
		    System.out.println("Trackball.java: Modifier: " + e.getModifiers());
		}
                isRightButton = (e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0;
            }
        });

	component.addMouseMotionListener(new MouseMotionAdapter() {
		public void mouseDragged(MouseEvent e) {
		    if(TTPproject.Ttp.commentslevel >= 3){
			System.out.println("Trackball.java: add mousedragged");
		    }
		    
		    //System.out.println("Mouse modifiers: " + 
		    //	       e.getModifiersExText(e.getModifiers()));
		    // all shift to TTPproject.CONSTANTS;
		    //int altdown_button1mask= InputEvent.ALT_MASK | InputEvent.BUTTON1_MASK;
		    //int ctrldown_button1mask= InputEvent.CTRL_MASK | InputEvent.BUTTON1_MASK;
 		    
		    //System.out.println("alt mask: " + InputEvent.ALT_MASK);
		    //System.out.println("Button1 mask: " + InputEvent.BUTTON1_MASK);
		    //System.out.println("getmodifier: " + e.getModifiers());
		    //System.out.println("ALT BUTTON 1 mask: " +  altdown_button1mask);
		    //System.out.println("CTRL BUTTON 1 mask: " +  ctrldown_button1mask);
		    
		    if((e.getModifiers() & ctrldown_button1mask) ==  ctrldown_button1mask){
			zooming3(e);
		    }
		    
		    
		    if((e.getModifiers() & altdown_button1mask) ==  altdown_button1mask){
			//System.out.println("Trackball.java: alt down + button 1: " + altdown_button1mask);
			rotate(e);
		    }
		    
		    if((e.getModifiers() & shiftdown_button1mask) ==  shiftdown_button1mask){
			//System.out.println("Trackball.java: alt down + button 1: " + altdown_button1mask);
			rotate(e);
		    }

		}
	    });

	component.addMouseWheelListener(this);

	component.addMouseMotionListener(new MouseMotionAdapter() {
		public void mouseMoved(MouseEvent e) {
		    if(TTPproject.Ttp.commentslevel >= 3){
			System.out.println("Trackball.java: mouse moved");
		    }

		    if(modifiers_keydown){
			if((e.getModifiers()) ==  ctrldown_mask){
			    if(first_time_moved){
				first_time_moved= false;
				v0 = makeTrackballVector(e);
				
			    }
			    zooming3(e);
			}
			
			
			if((e.getModifiers()) ==  altdown_mask){
			    //System.out.println("Trackball.java: alt down + button 1: " + altdown_button1mask);
			    if(first_time_moved){
				first_time_moved= false;
				v0 = makeTrackballVector(e);
			    }
			    rotate(e);
			}
			
			if((e.getModifiers()) ==  shiftdown_mask){
			    //System.out.println("Trackball.java: alt down + button 1: " + altdown_button1mask);
			    if(first_time_moved){
				first_time_moved= false;
				v0 = makeTrackballVector(e);
			    }
			    rotate(e);
			}
		    }

		    
		}
		
		
	    });

	
	component.addKeyListener(new KeyAdapter(){
		public void keyPressed(KeyEvent e){
		    if(TTPproject.Ttp.commentslevel >= 3){
			System.out.println("Scene.java: add keyPressed");
		    }
		    
		    if(e.isControlDown() || e.isAltDown() || e.isShiftDown()){// modifiers keyed up
			modifiers_keydown= true;
			
		    }
		    
		} 
	    });


	

	component.addKeyListener(new KeyAdapter(){
		public void keyReleased(KeyEvent evt){
		    if(TTPproject.Ttp.commentslevel >= 2){
			System.out.println("Trackball.java: add keyReleased");
		    }

		    if(modifiers_keydown){
			modifiers_keydown= false;
			first_time_moved= true;
		    }
		    		    
		} 
	    });

	

	
	this.minZoom = minZoom;
        this.maxZoom = maxZoom;
	this.config= config;
	zoomFactor = (float)config.getLowerZoomfactor();
	lengthFactor= 1.f;
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: Trackball() initial transformation:  " 
			       + totalRotation.toString() );
	}
	
    }


    public void mouseWheelMoved(MouseWheelEvent e)
    {
	if(TTPproject.Ttp.commentslevel >= 2){
	    System.out.println("Trackball.java: wheel moved");
	}
	
	int scroll_amount= e.getScrollAmount();
	int scroll_unit= e.getUnitsToScroll();
	int rotations= e.getWheelRotation();
	int scroll_type= e.getScrollType();
	if(scroll_type == MouseWheelEvent.WHEEL_UNIT_SCROLL ){
	    if(TTPproject.Ttp.commentslevel >= 3){
		System.out.println("Trackball.java: UNIT");
	    }
	    
	    //zooming2(scroll_unit);
	    zooming3(scroll_unit);
	}

	else if(scroll_type == MouseWheelEvent.WHEEL_BLOCK_SCROLL){
	    if(TTPproject.Ttp.commentslevel >= 3){
		System.out.println("Trackball.java: BLOCK");
	    }
	}
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: Scroll amount= " + scroll_amount);
	    System.out.println("Trackball.java: wheel rotations= " + rotations);
	    System.out.println("Trackball.java: Scroll unit= " + scroll_unit);
	}
	
    }

/** Returns the current rotation matrix in OpenGL format. The matrix is
 *  returned in a 1D float array, columnwise, i.e. the elements in
 *  column1, followed by the elements in column2, etc.
 */
    public float[] getGLRotation() {
        return totalRotation.getGLMatrix();
    }

    public void setIdentity()
    {
	totalRotation.setIdentity();
    }

    /*public void reset()
    {
	setIdentity();
	config.lengthfactor= 1.f;
	lengthFactor= 1.f;
	}*/
    

/** Returns the current rotation matrix in geometry.Transformation form.
 */
    public Transformation getRotation() {
       return totalRotation;
    }

/** Returns a rotation matrix in OpenGL format that performs the appropriate
 *  combination of an OpenGL gluLookAt call followed by the trackball rotation.
 *  Trackball rotation is about the lookAt point. 
 *  Because the trackball rotation is applied by a user to an object seen in
 *  view space, the trackball rotation must be applied after the view rotation.
 *  However, the view rotation must be applied after the origin has been moved
 *  to the eyepoint, and this is not the correct rotation point for the trackball!
 *  There are thus five steps:
 *  (1) translate scene so origin is at eye point;
 *  (2) apply view rotation;
 *  (3) translate scene so origin is at the (original) look at point
 *  (4) apply trackball rotation
 *  (5) put origin back to where it was after step 2.
 *  
 *  @param xEye The x coordinate of the eye position
 *  @param yEye The y coordinate of the eye position
 *  @param zEye The z coordinate of the eye position
 *  @param xLookAt The x coordinate of the LookAt position
 *  @param yLookAt The y coordinate of the LookAt position
 *  @param zLookAt The z coordinate of the LookAt position
 *  @param xUp The x coordinate of the up vector
 *  @param yUp The y coordinate of the up vector
 *  @param zUp The z coordinate of the up vector
 */
    /*public float[] getLookAtPlusTrackballRotation(double xEye, double yEye, double zEye,
						  double xLook, double yLook, double zLook,
						  double xUp, double yUp, double zUp) {
	Vector3d nEyeToLookAt = new Point3d(xEye, yEye, zEye).minus(new Point3d(xLook, yLook, zLook));
	
	Vector3d n = nEyeToLookAt.normalised();
	Vector3d up = new Vector3d(xUp, yUp, zUp);
	Vector3d u = up.cross(n).normalised();
	Vector3d v = n.cross(u);
	double dEyeToLookAt = nEyeToLookAt.length();
	return Transformation.shift(-xEye, -yEye, -zEye).then(   // Eye to origin
	       Transformation.alignRotation(u, v, n)).then(        // Align eye axes with world
	       Transformation.shift(0,0,dEyeToLookAt)).then(       // Orig. lookat to origin
	       totalRotation).then(                                // Trackball rotation
	       Transformation.shift(0,0,-dEyeToLookAt)).getGLMatrix();  // Restore origin
	       }*/


    public Transformation getLookAtPlusTrackballRotation(double xEye, double yEye, double zEye,
							 double xLook, double yLook, double zLook,
							 double xUp, double yUp, double zUp) {
	Vector3d nEyeToLookAt = new Point3d(xEye, yEye, zEye).minus(new Point3d(xLook, yLook, zLook));
	
	Vector3d n = nEyeToLookAt.normalised();
	Vector3d up = new Vector3d(xUp, yUp, zUp);
	Vector3d u = up.cross(n).normalised();
	Vector3d v = n.cross(u);
	double dEyeToLookAt = nEyeToLookAt.length();
	return Transformation.shift(-xEye, -yEye, -zEye).then(   // Eye to origin
	       Transformation.alignRotation(u, v, n)).then(        // Align eye axes with world
	       Transformation.shift(0,0,dEyeToLookAt)).then(       // Orig. lookat to origin
	       totalRotation).then(                                // Trackball rotation
	       Transformation.shift(0,0,-dEyeToLookAt));  // Restore origin
    }

    public float[] getLookAt(double xEye, double yEye, double zEye,
			     double xLook, double yLook, double zLook,
			     double xUp, double yUp, double zUp) {
	Vector3d nEyeToLookAt = new Point3d(xEye, yEye, zEye).minus(new Point3d(xLook, yLook, zLook));
	
	Vector3d n = nEyeToLookAt.normalised();
	Vector3d up = new Vector3d(xUp, yUp, zUp);
	Vector3d u = up.cross(n).normalised();
	Vector3d v = n.cross(u);
	double dEyeToLookAt = nEyeToLookAt.length();
	
	return Transformation.shift(-xEye, -yEye, -zEye).then(   // Eye to origin
	       Transformation.alignRotation(u, v, n)).then(      // Align eye axes with world
	       Transformation.shift(0,0,dEyeToLookAt)).then(     // Orig. lookat to origin
	       Transformation.shift(0,0,-dEyeToLookAt)).getGLMatrix();  // Restore origin
    }
    

    
/** Returns the current zoom factor. The factor is guaranteed to be in
 *  the range minZoom to maxZoom as defined by the default values
 *  or explicitly in the constructor.
 *  @see #MIN_ZOOM_DEFAULT
 *  @see #MAX_ZOOM_DEFAULT
 */
    public float getZoomFactor() {
        return zoomFactor;
    }

/** Enables/disables calls to the parent component's repaint method.
 * @param repaintReqd true/false to enable/disable calls to the parent
 * component's repaint method. Default is true.
 */
    public void setRepaintReqd(boolean repaintReqd) {
       this.repaintReqd = repaintReqd;
    }

    
    private void rotate(MouseEvent e) {
	Vector3f v1 = makeTrackballVector(e);

	Vector3f axis = v0.cross(v1);
	if (axis.length() > 1.0e-4f) { // Ignore negligible drags
	    axis = axis.normalised();
	    float dotProd = Math.min(1, v0.dot(v1));
	    float degrees = (float) (Math.acos(Math.min(1, v0.dot(v1))) * 180 / Math.PI);
	    totalRotation.multiplyOnLeft(Transformation.rotation(degrees, axis));
	    
	    if(TTPproject.Ttp.commentslevel >= 3){
		System.out.println("degrees: " + degrees);
	    }
	}
	v0 = v1;
	if (repaintReqd)
	    component.repaint();
	
    }

    private void zooming(MouseEvent e){
	Vector3f v1 = makeTrackballVector(e);
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: zooming() delta y = " + (v1.dy() - v0.dy()));
	}
	
	zoomFactor *= (1 + (v1.dy() - v0.dy()) * ZOOM_RATE);
	zoomFactor = Math.max(minZoom, Math.min(maxZoom, zoomFactor));
	v0 = v1;
        if (repaintReqd)
	    component.repaint();
    }

    public void zooming2(int unit)
    {
	if(TTPproject.Ttp.commentslevel >= 2){
	    System.out.println("Trackball.java: zooming2()");
	}
	
	float deltaY= (float)-unit * 0.05f;
	
	zoomFactor *= (1 + (deltaY)* ZOOM_RATE);
	zoomFactor = Math.max(minZoom, Math.min(maxZoom, zoomFactor));
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: zooming2() modified zoomFactor= " + zoomFactor);
	}

	if (repaintReqd)
	    component.repaint();

	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: unit> " + unit);
	    System.out.println("Trackball.java: deltaY> " + deltaY);
	    System.out.println("Trackball.java: zoomFactor> " + zoomFactor);
	}
    }
    
    /*
     * change lengthFactor
    */
    public void zooming3(int unit)
    {
	if(TTPproject.Ttp.commentslevel >= 2){
	    System.out.println("Trackball.java: zooming3(unit)");
	}
	
	float delta= (float)unit * 0.05f;

	config.lengthfactor *= (1 + (delta)* ZOOM_RATE);
	config.lengthfactor= Math.max(config.lower_length_factor, Math.min( config.upper_length_factor , config.lengthfactor));
	
	//lengthFactor = Math.max(minZoom, Math.min(maxZoom, zoomFactor));// yet to be changed
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: zooming3() modified zoomFactor= " + zoomFactor);
	}

	if (repaintReqd)
	    component.repaint();
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: unit> " + unit);
	    System.out.println("Trackball.java: deltaY> " + delta);
	    System.out.println("Trackball.java: lengthFactor> " + config.lengthfactor);
	}
	
    }

    private void zooming3(MouseEvent e)
    {
	if(TTPproject.Ttp.commentslevel >= 2){
	    System.out.println("Trackball.java: zooming3(mouseevent)");
	}
	Vector3f v1 = makeTrackballVector(e);

	config.lengthfactor *= (1 + (v0.dy() - v1.dy()) * ZOOM_RATE);
	
	config.lengthfactor= Math.max(config.lower_length_factor, Math.min( config.upper_length_factor , config.lengthfactor));

	v0 = v1;
	
	//lengthFactor = Math.max(minZoom, Math.min(maxZoom, zoomFactor));// yet to be changed
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: zooming3() modified zoomFactor= " + zoomFactor);
	}

	if (repaintReqd)
	    component.repaint();

	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Trackball.java: lengthFactor> " + config.lengthfactor);
	}
    }
    
    private void handleDrag(MouseEvent e) {
        // Handle mouse drags. Called directly by MouseMotionAdaptor.
	//System.out.println("Trackball.java: handledDrag()");
	
        Vector3f v1 = makeTrackballVector(e);
        if (isRightButton)  {
            zoomFactor *= (1 + (v1.dy() - v0.dy()) * ZOOM_RATE);
            zoomFactor = Math.max(minZoom, Math.min(maxZoom, zoomFactor));
        }
        else  {
            Vector3f axis = v0.cross(v1);
            if (axis.length() > 1.0e-4f) { // Ignore negligible drags
                axis = axis.normalised();
                float dotProd = Math.min(1, v0.dot(v1));
                float degrees = (float) (Math.acos(Math.min(1, v0.dot(v1))) * 180 / Math.PI);
                totalRotation.multiplyOnLeft(Transformation.rotation(degrees, axis));
		//System.out.println("degrees: " + degrees);
            }
        }
        v0 = v1;
        if (repaintReqd)
	    component.repaint();
    }

    private Vector3f makeTrackballVector(MouseEvent e) {
        Point p = e.getPoint();
        Dimension dim = e.getComponent().getSize();
        int width = dim.width;
        int height = dim.height;
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Dimension:( " + width + ", " +
			       height + " )");
	    System.out.println("p:( " + p.x + ", " +
			       p.y + " )");
	}
	
        float x = ( 2 * p.x - width ) / (float) width;
        float y = ( height - 2 * p.y ) / (float) height;
        float tmp = 1 - x*x - y*y;
        float z = tmp <= 0 ? 0 : (float) Math.sqrt(tmp);
	
	if(TTPproject.Ttp.commentslevel >= 3){
	    System.out.println("Vector:( " + x + ", " +
			       y + ", "+
			       z + " )");
	}

        return new Vector3f(x, y, z).normalised();
    }
}
