package geometry;

 
/**
 * Represents a linear transformation that can be applied a homogeneous-coordinate point.

 * A transformation is an operator that is considered to be applied to
 * a point ON THE LEFT, i.e. p' = T p, where p is
 * a point that is transformed by T to p'. A Transformation is represented internally
 * as a 4 x 4 matrix, and this implementation assumption pervades the documentation
 * descriptions.
 *
 * @author Richard Lobb
 * @version 1.2, May 2002
 */

/* Change history:
 * Modified to use doubles internally rather than floats, and to interface to
 * both Vector3f's and Vector3d's.
 * May 2002.  RJL
 */

public class Transformation {

/**
 * Symbolic constants denoting the three coordinate axes
 */
    public static final int X_AXIS = 0, Y_AXIS = 1, Z_AXIS = 2;

/**
 * The transformation in 4 x 4 matrix form.
 */
    protected double[][] m = new double[4][4];
    

    private float[] glMatrix = null;
 
    // Constructors
    // ============

/**
 * Constructs the identity transformation
 */
    public Transformation() {
	setIdentity();
    }

    public void setIdentity()
    {
	// Parameterless constructor defines identity matrix
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                m[i][j] = (i == j ? 1 : 0);
    }

/**
 * Constructs a transformation as defined by a given 4 x 4 matrix of floats.
 * The matrix is assumed to multiply a point (represented as a four-element
 * column vector) on the left.
 */
    public Transformation(float[][] m) {
        if (m.length != 4)
            throw new RuntimeException("Transformation: must be given a 4 x 4 array");
        for (int i = 0; i < 4; i++) {
            if (m[i].length != 4)
                throw new RuntimeException("Transformation: must be given a 4 x 4 array");
            for (int j = 0; j < 4; j++)
                this.m[i][j] = m[i][j];
        }
    }

/**
 * Constructs a transformation as defined by a given 4 x 4 matrix of doubles.
 * The matrix is assumed to multiply a point (represented as a four-element
 * column vector) on the left.
 */
    public Transformation(double[][] m) {
        if (m.length != 4)
            throw new RuntimeException("Transformation: must be given a 4 x 4 array");
        for (int i = 0; i < 4; i++) {
            if (m[i].length != 4)
                throw new RuntimeException("Transformation: must be given a 4 x 4 array");
            for (int j = 0; j < 4; j++)
                this.m[i][j] = m[i][j];
        }
    }
/**
 * Constructs a transformation by copying (cloning) an existing transformation.
 * @param t The transformation to be copied.
 */
    public Transformation(Transformation t) {
        // Copy constructor
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                this.m[i][j] = t.m[i][j];
    }

 /**
  * Constructs a transformation representing a translation by a given Vector3f.
  * @param v The translation vector.
  */
   public Transformation(Vector3f v) {
        this();
        m[0][3] = v.dx(); m[1][3] = v.dy(); m[2][3] = v.dz();
    }

/**
  * Constructs a transformation representing a translation by a given Vector3d.
  * @param v The translation vector.
  */
   public Transformation(Vector3d v) {
        this();
        m[0][3] = v.dx(); m[1][3] = v.dy(); m[2][3] = v.dz();
    }

 /**
  * Constructs a transformation representing a uniform scaling by a given scale factor.
  * @param scale The scale factor along the three axes.
  */
    public Transformation(double scale) {
        // [The case of Transformation(float, float, float) is not
        // provided because of possible confusion with translation]
        this();
        m[0][0] = m[1][1] = m[2][2] = scale;
    }


    // Static methods to return specific transformations
    // =================================================

 /**
  * A transformation representing a translation by a given vector.
  * @param v The translation vector.
  * @return A transformation that translates any point to which it is applied by "v".
  */
    public static Transformation shift(Vector3f v) {
        return new Transformation(v);
    }

/**
  * A transformation representing a translation by a given vector.
  * @param v The translation vector.
  * @return A transformation that translates any point to which it is applied by "v".
  */
    public static Transformation shift(Vector3d v) {
        return new Transformation(v);
    }
    
 /**
  * A transformation representing a translation by a given vector.
  * @param dx The translation in the x direction.
  * @param dy The translation in the y direction. 
  * @param dz The translation in the z direction.
  * @return A transformation that translates any point to which it is applied by (dx,dy,dz).
  */
    public static Transformation shift(double dx, double dy, double dz) {
        // A translation by (dx, dy, dz)
        return new Transformation(new Vector3d(dx, dy, dz));
    }

 /**
  * A transformation representing a scaling by given factors along the three axes.
  * @param xs The scale factor in the x direction.
  * @param ys The scale factor in the y direction. 
  * @param zs The scale factor in the z direction.
  * @return A transformation that scales any point to which it is applied by (xs,ys,zs).
  */
    public static Transformation scaling(double xs, double ys, double zs) {
        Transformation tx = new Transformation();
        tx.m[0][0] = xs;
        tx.m[1][1] = ys;
        tx.m[2][2] = zs;
        return tx;
    }

 /**
  * A transformation representing a right-handed 
  * rotation through a given angle (in degrees) around
  * a given axis vector (type Vector3d) through the origin.
  * @param degrees The angle in degrees (measured in a right-handed sense
  * with respect to the rotation axis)
  * @param axis The axis of rotation.
  * @return A transformation that rotates any point to which it is applied 
  * through an angle of "AngleInDegrees" around the given axis.
  */
    public static Transformation rotation(double angleInDegrees, Vector3d axis) {
        double x = axis.dx(), y = axis.dy(), z = axis.dz();
        double theta = Math.PI * angleInDegrees / 180;
        double c = Math.cos(theta);
        double s = Math.sin(theta);
        double t = 1 - c;
        double[][] m = {
            {t*x*x+c, t*x*y-s*z, t*x*z+s*y, 0},
            {t*x*y+s*z, t*y*y+c, t*y*z-s*x, 0},
            {t*x*z-s*y, t*y*z+s*x, t*z*z+c, 0},
            {0,0,0,1}};
        return new Transformation(m);
    }

/**
  * A transformation representing a right-handed 
  * rotation through a given angle (in degrees) around
  * a given axis vector (of type Vector3f) through the origin.
  * @param degrees The angle in degrees (measured in a right-handed sense
  * with respect to the rotation axis)
  * @param axis The axis of rotation.
  * @return A transformation that rotates any point to which it is applied 
  * through an angle of "AngleInDegrees" around the given axis.
  */
    public static Transformation rotation(double angleInDegrees, Vector3f axis) {
        return rotation(angleInDegrees, new Vector3d(axis.dx(), axis.dy(), axis.dz()));
    }
    

 /**
  * A transformation that rotates the world coordinate system to align with
  * a given coordinate system specified with Vector3d's.
  * @param u A unit vector along the "x" axis of the new coordinate system
  * @param v A unit vector along the "y" axis of the new coordinate system 
  * @param n A unit vector along the "z" axis of the new coordinate system 
  * @return A rotation transformation that represents a change of coordinate
  * system from the (x,y,z) system to the (u,v,n) system with the same origin.
  */
   public static Transformation alignRotation(Vector3d u, Vector3d v, Vector3d n) {
        // Returns the transformation that rotates a (u,v,n) coordinate
        // system to align with world axes.
        double[][] m = {
            {u.dx(), u.dy(), u.dz(), 0},
            {v.dx(), v.dy(), v.dz(), 0},
            {n.dx(), n.dy(), n.dz(), 0},
            {0, 0, 0, 1}};
        return new Transformation(m);
    }

/**
  * A transformation that rotates the world coordinate system to align with
  * a given coordinate system specified with Vector3f's.
  * @param u A unit vector along the "x" axis of the new coordinate system
  * @param v A unit vector along the "y" axis of the new coordinate system 
  * @param n A unit vector along the "z" axis of the new coordinate system 
  * @return A rotation transformation that represents a change of coordinate
  * system from the (x,y,z) system to the (u,v,n) system with the same origin.
  */
   public static Transformation alignRotation(Vector3f u, Vector3f v, Vector3f n) {
        // Returns the transformation that rotates a (u,v,n) coordinate
        // system to align with world axes.
        double[][] m = {
            {u.dx(), u.dy(), u.dz(), 0},
            {v.dx(), v.dy(), v.dz(), 0},
            {n.dx(), n.dy(), n.dz(), 0},
            {0, 0, 0, 1}};
        return new Transformation(m);
    }

    // Functional-style (immutable) operations
    // =======================================


/**
  * A transformation "this" * m. Returns a new transformation that is
  * the matrix product of this times another matrix m. Since the
  * transformation matrices muliply points on the left, the effect
  * is that the new transformation is the result of applying the
  * given transformation first then "this".
  * @param m The other transformation.
  * @return The transformation "this" * m.
  * @see #then
  */
    public Transformation times(Transformation m) {
        double[][] result = new double[4][4];
        for (int i = 0; i < 4; i++)  {
            for (int j = 0; j < 4; j++) {
                double sum = 0;
                for (int k = 0; k < 4; k++)
                    sum += this.m[i][k] * m.m[k][j];
                result[i][j] = sum;
            }
        }
        return new Transformation(result);
    }

/**
  * A transformation m * "this". Returns a new transformation that is
  * the matrix product of this m * "this". Since the
  * transformation matrices muliply points on the left, the effect
  * is that the new transformation is the result of applying "this"
  * then the given transformation.
  * @param m The given transformation.
  * @return The transformation m * "this"
  * @see #times
  */
    public Transformation then(Transformation m) {
         return m.times(this);
    }

/**
 * The result of transforming a given Point3d by "this".
 * @param p The point to be transformed
 * @return The transformed point "this" * p.
 */
    public Point3d times(Point3d p) {
        return p.transformedBy(this);
    }

/**
 * The result of transforming a given Point3f by "this".
 * @param p The point to be transformed
 * @return The transformed point "this" * p.
 */
    public Point3f times(Point3f p) {
        return p.transformedBy(this);
    }

 /**
 * The result of transforming a given Vector3f by "this".
 * DANGER! USE WITH CAUTION! When vectors are transformed
 * with this method, the translation component of the transformation
 * is ignored. This method should not be used to transform either
 * position vectors of points or surface normal vectors (which should
 * be transformed by the inverse transpose of the transformation).
 * @param p The point to be transformed
 * @return The transformed point "this" * p.
 */
   public Vector3f times(Vector3f v) {
        return v.transformedBy(this);
    }

/**
 * The result of transforming a given Vector3d by "this".
 * DANGER! USE WITH CAUTION! When vectors are transformed
 * with this method, the translation component of the transformation
 * is ignored. This method should not be used to transform either
 * position vectors of points or surface normal vectors (which should
 * be transformed by the inverse transpose of the transformation).
 * @param p The point to be transformed
 * @return The transformed point "this" * p.
 */
   public Vector3d times(Vector3d v) {
        return v.transformedBy(this);
    }


/**
 * The result of applying this transformation to a given point.
 * @param p The point to which the transformation is applied
 * @return The point p after transformation by "this"
 * @deprecated Use "times" instead.
 */
    public Point3f appliedTo(Point3f p) {
        // Return the point p after transformation by this.
        // DEPRECATED. USE TIMES. THIS IS LEGACY CODE FOR 715 USE ONLY
        return p.transformedBy(this);
    }

/**
 * The inverse of "this".
 * This implementation uses a non-robust algorithm but should be good enough for
 * normal use (?).The algorithm is: extract the translation (row 4), use an explicit
 * formula for the inverse of the upper-left 3 x 3 matrix, and then
 * multiply by the inverse of the extracted translation.
 * This algorithm only works with transformations that are "well formed", meaning
 * that the fourth row is (0,0,0,1).
 * @return The inverse of "this".
 * @throw RuntimeException if the transformation is not "well formed" or is ill-conditioned.
 */
    public Transformation inverse() {
        for (int i = 0; i < 4; i++)
            if (Math.abs(m[3][i] - (i == 3 ? 1 : 0)) > 0.00001)
                throw new RuntimeException("Transformation.inverse: ill-formed matrix");
            double a = m[0][0],
                b = m[0][1],
                c = m[0][2],
                d = m[1][0],
                e = m[1][1],
                f = m[1][2],
                g = m[2][0],
                h = m[2][1],
                i = m[2][2];
            double determinant = -c*e*g+b*f*g+c*d*h-a*f*h-b*d*i+a*e*i;
            if (Math.abs(determinant) < 0.00001)
                throw new RuntimeException("Transformation.inverse: ill-conditioned matrix");
            double[][] m2 = {{-f*h+e*i,c*h-b*i,-c*e+b*f,0},
                            {f*g-d*i,-c*g+a*i,c*d-a*f,0},
                            {-e*g+d*h,b*g-a*h,-b*d+a*e,0},
                            {0,0,0,1}};
            for (int ii = 0; ii < 3; ii++)
                for (int jj = 0; jj < 3; jj++)
                    m2[ii][jj] /= determinant;
                return shift(-m[0][3],-m[1][3],-m[2][3]).then(new Transformation(m2));
    }

/**
 * The transpose of "this". WARNING! This can be a trap -- it does NOT transpose
 * this in place, but rather returns a new transformation, leaving "this" unchanged.
 * @return The transpose of "this", i.e. the transformation (regarded as a matrix) whose
 * rows are equal to the columns of "this".
 */
    public Transformation transpose() {
        // The transpose of "this".
        double[][] m2 = {{m[0][0], m[1][0], m[2][0], m[3][0]},
            {m[0][1], m[1][1], m[2][1], m[3][1]},
            {m[0][2], m[1][2], m[2][2], m[3][2]},
            {m[0][3], m[1][3], m[2][3], m[3][3]}};
        return new Transformation(m2);
    }
    
/**
 * The matrix representation of "this".
 * @return A 4 x 4 float array representing "this" (i.e.
 * a copy of the protected instance variable m).
 */
    public float[][] matrixf() {
        float[][] result = new float[4][4];
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                result[i][j] = (float) this.m[i][j];
        return result;
    }

/**
 * The matrix representation of "this".
 * @return A 4 x 4 double array representing "this" (i.e.
 * a copy of the protected instance variable m).
 */
    public double[][] matrixd() {
        double[][] result = new double[4][4];
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                result[i][j] = this.m[i][j];
        return result;
    }

 /**
  * The matrix representation of "this" in the strange transposed
  * form suitable for use with OpenGL.
  * @return A 4 x 4 float array representing "this" in OpenGL form, 
  * i.e. a transpose of the internal representation, m.
  */
   public float[] getGLMatrix() {
        // Return the matrix columnwise, for OpenGL use
        if (glMatrix == null)
            glMatrix = new float[] 
            {(float) m[0][0], (float) m[1][0], (float) m[2][0], (float) m[3][0],
             (float) m[0][1], (float) m[1][1], (float) m[2][1], (float) m[3][1],
             (float) m[0][2], (float) m[1][2], (float) m[2][2], (float) m[3][2],
             (float) m[0][3], (float) m[1][3], (float) m[2][3], (float) m[3][3]};
        return glMatrix;
    }

/**
 * A string represention of the transformation
 * @return A string representation "{{m00, m01, m02, m03}, {m10, ..... }}"
 * of the matrix form of the transformation.
 */
    public String toString() {
        String result = "{";
        for (int row = 0; row < 4; row++) {
            result += "{";
            for (int col = 0; col < 4; col++) {
                result += m[row][col];
                if (col != 3)
                    result += ",";
            }
            result += "}";
        }
        return result + "}";
    }

    // Non-functional (mutable) operations
    // ===================================
    // I implement 'em as I need 'em.
    // Haven't needed many yet!


/**
 * Multiplies "this" on the left by a given transformation.
 * @param mLeft The transformation to be multiplied with this.
 * Sets the internal matrix representation of the transformation
 * "this" to mLeft * this.
 */
    public void multiplyOnLeft(Transformation mLeft) {
        double[][] newM = new double[4][4];
        for (int i = 0; i < 4; i++) {
            for (int j = 0; j < 4; j++) {
                double sum = 0;
                for (int k = 0; k < 4; k++)
                    sum += mLeft.m[i][k] * m[k][j];
                newM[i][j] = sum;
            }
        }
        m = newM;
        glMatrix = null;
    }
}

