package geometry;

/**
 * A vector in double precision floating-point 3 space.
 * @author Richard Lobb
 * @version 1.0, May 2002
 */
 
 public class Vector3d {

/**
 * The x component of the vector.
 */
    protected double dx;
/**
 * The y component of the vector.
 */
    protected double dy;
/**
 * The z component of the vector.
 */
    protected double dz;

    // Constructors
    // ============

/**
 * Constructs a zero vector (0,0,0)
 */
    public Vector3d() {
        dx = dy = dz = 0;
    }
  
/**
 * Constructs a vector (dx, dy, dz)
 * @param dx The x-component of the vector
 * @param dy The y-component of the vector
 * @param dz The z-component of the vector 
 */
    public Vector3d(double dx, double dy, double dz) {

    this.dx = dx; this.dy = dy; this.dz = dz;
    }
        
/**
 * Constructs a copy of a given vector
 * @param v The vector to be copied ("cloned").
 */
    public Vector3d(Vector3d v) {
        dx = v.dx; dy = v.dy; dz = v.dz;
    }
        
/**
 * Constructs the position vector of a given point
 * @param p The point whose position vector is to be constructed.
 */
    public Vector3d(Point3d p) {
        dx = p.x(); dy = p.y(); dz = p.z();
    }
  
/**
 * Constructs a vector from a 3-element double array.
 * @param components The 3-element array of doubles to use
 * as components of the vector.
 * @throws RuntimeException if the given array is not a 3-element array.
 */
    public Vector3d(double[] components) {
        if (components.length == 3) {
            dx = components[0];
            dy = components[1];
            dz = components[2];
        }
        else 
            throw new RuntimeException("Vector3d: constructor needs exactly 3 doubles");
    }

  // Component access
  
 /**
 * The <i>x</i> component of the vector
 */
    public double dx() { return dx; }
/**
 * The <i>y</i> component of the vector
 */
    public double dy() { return dy; }
/**
 * The <i>z</i> component of the vector
 */
    public double dz() { return dz; }

 
  /**
  * True if and only if all components of "this" are equal to the corresponding
  * components of another vector to within a specified tolerance.
  * @param v The other vector
  * @param tolerance The maximum permitted absolute difference between any component of "this"
  * and the corresponding component of "v" for the result returned to be true.
  * @return True if and only if all components of "this" are equal to the corresponding
  * components of <i>v</i> to within the specified tolerance.
  */
    public boolean epsilonEquals(Vector3d v, double tolerance) {
        return
          Math.abs(dx - v.dx) <= tolerance &&
          Math.abs(dy - v.dy) <= tolerance &&
          Math.abs(dz - v.dz) <= tolerance;
    }

  
 /**
 * A 3-element float vector representing "this"
 * @return A new double array {dx, dy, dz}. This form
 * is useful for passing to the OpenGL engine.
 */
   public float[] get3fv() {
        return new float[] { (float) dx, (float) dy, (float) dz };
    }
  
/**
 * Copies the components of "this" into the given double array.
 * @param p the target array. This must be an array of
 * 3 or more doubles. Copies the x, y and z components of the
 * vector into <code>p[0], p[1], p[2]</code> respectively.
 * @throws RuntimeException if the array length is less than 3.
*/
    public void copyInto(double[] v) {
        if (v.length < 3)
            throw new RuntimeException("Vector3d.copyInto: target array length < 3");
        v[0] = dx; v[1] = dy; v[2] = dz;
    }

/**
  * A string represention of "this".
  * @return A string represention of "this", such as "(2.001, 3, 4.1)".
  * The individual components are of whatever length the standard Java
  * double-to-string conversion routines deem appropriate
  */
    public String toString() {
        return "(" + dx() + "," + dy() + "," + dz() + ")";
    }
 
 // Functional style arithmetic routines.
 // =====================================
 
 /**
 * A new vector equal to "this" plus another vector.
 * @param v the vector to be added to "this"
 * @return a new vector equal to "this" + v
 */
    public Vector3d plus(Vector3d v) {
        return new Vector3d(dx+v.dx, dy+v.dy, dz+v.dz);
    }

 /**
 * A new vector equal to "this" minus another vector.
 * @param v The vector to be added to "this"
 * @return A new vector equal to "this" - v
 */
    public Vector3d minus(Vector3d v) {
        return new Vector3d(dx-v.dx, dy-v.dy, dz-v.dz);
    }

 /**
 * A new vector equal to "this" times a scale factor
 * @param scale the scale factor
 * @return a new Vector3d object equal to "this" * scale
 */
    public Vector3d times(double scale) {
        return new Vector3d(dx*scale, dy*scale, dz*scale);
    }
    
 /**
  * A new vector equal to "this" negated
  * @return a new vector equal to "this" with all components negated.
  */
    public Vector3d negated() {
        return new Vector3d(-dx, -dy, -dz);
    }   
  
/**
  * A new vector equal to "this" with its length scaled to 1.
  * @return a new vector with the same direction as "this" but with unit length.
  */
    public Vector3d normalised() {
        return this.times(1/this.length());
    }
 
 /**
  * The dot product of "this" and a given vector.
  * @param v The other vector
  * @return The dot product of "this" and v.
  */
    public double dot(Vector3d v) { 
        return dx * v.dx + dy * v.dy + dz * v.dz;
    }
    
 /**
  * The length of "this"
  * @return The length of "this"
  */
    public double length() {
        return Math.sqrt(dx*dx + dy*dy + dz*dz);
    }
        
/**
  * The cross product of "this" and a given vector.
  * @param v The other vector
  * @return A new vector equal to the cross product of "this" and v.
  */
    public Vector3d cross(Vector3d v) { 
        return new Vector3d(
            dy*v.dz - dz*v.dy,
            dz*v.dx - dx*v.dz,
            dx*v.dy - dy*v.dx);
    }
    
/**
  * A new vector equal to "this" when transformed by a given transformation.
  * NB: Since "this" is a vector, translations have no effect. DO NOT USE on
  * position vectors or surface-normal vectors.
  * @param trans The transformation
  * @return A new vector equal to the result of transforming "this"
  * by the transformation "trans".
  */
    public Vector3d transformedBy(Transformation trans) {
        double[][] m = trans.matrixd();
        Vector3d result = new Vector3d();
        result.dx = dx * m[0][0] + dy * m[0][1] + dz * m[0][2];
        result.dy = dx * trans.m[1][0] + dy * m[1][1] + dz * m[1][2];
        result.dz = dx * trans.m[2][0] + dy * m[2][1] + dz * m[2][2];
        return result;
    }


 /**
  * The angle in degrees between "this" and another vector.
  * @param v The other vector
  * @return The angle in degrees between "this" and <i>v</i>.
  */
    public double AngleInDegs(Vector3d v) {
        return Math.acos(this.normalised().dot(v.normalised())) * 180/Math.PI;
    }
 
    
  // In-place style arithmetic routines.[i.e. the "mutables"]
  // ========================================================
  
/**
 * Sets the components of "this" to be equal to those of a given vector
 * @param v the vector whose components are to be copied.
 */
    public void set(Vector3d v) {
        this.dx = v.dx;
        this.dy = v.dy;
        this.dz = v.dz;
    }       
    
/**
 * Sets the components of "this" to the given values.
 * @param dx the x component
 * @param dy the y component
 * @param dz the z component
 */
    public void set(double dx, double dy, double dz) {
        this.dx = dx;
        this.dy = dy;
        this.dz = dz;
    }
    
      
/**
 * Sets "this" to the result of scaling a given vector by a given amount.
 * The value of "this" is set to <code>scale * v</code>.
 * @param scale The scale factor.
 * @param v The vector to be scaled.
 */
    public void setScaled(double scale, Vector3d v) {
        dx = scale * v.dx; dy = scale * v.dy; dz = scale * v.dz;
    }
    
/**
 * Adds a vector to "this". The value of "this" is set to "this" + v.
 * @param v The vector to be added to "this".
 */
    public void add(Vector3d v) {
        dx += v.dx; dy += v.dy; dz += v.dz;
    }
  
/**
 * Subracts a vector from "this". The value of "this" is set to "this" - v.
 * @param v The vector to be subtracted from "this".
 */
    public void subtract(Vector3d v) {
        dx -= v.dx; dy -= v.dy; dz -= v.dz;
    }
  
/** 
 * Adds the position vector of a given point to "this".
 * @param p The point whose position vector is to be added to "this".
 */
    public void add(Point3d p) {
        dx += p.x(); dy += p.y(); dz += p.z();
    }
  
/** 
 * Subtracts the position vector of a given point from "this".
 * @param p The point whose position vector is to be subtracted from "this".
 */
    public void subtract(Point3d p) {
        dx -= p.x(); dy -= p.y(); dz -= p.z();
    }
  
 /**
  * Multiplies "this" by a given scale factor. "This" is set to "this" * scale.
  * @param scale The scale factor by which all components of "this" are to be multiplied.
  */
    public void multiply(double scale) {
        dx *= scale;
        dy *= scale;
        dz *= scale;
    }
  
 /**
  * Scales "this" to a length of 1.
  */
    public void normalise() {
        multiply(1 / length());
    }
 
 /**
  * Adds to this a given vector scaled by a given amount.
  * @param scale The scale factor
  * @param The vector that is scaled by "scale" before being added to "this".
  */  
    public void addScaled(double scale, Vector3d v) {
    // Sets this = this + scale * v;
        dx += scale * v.dx;
        dy += scale * v.dy;
        dz += scale * v.dz;
    }
  
 /**
  * Performs the cross product of "this" and another vector.
  * Sets this to have the value "this" cross v.
  * @param v The other vector.
  */  
    public void crossWith(Vector3d v) {
      // Set this = this cross v
        double   c0 = dy*v.dz - dz*v.dy,
                 c1 = dz*v.dx - dx*v.dz,
                 c2 = dx*v.dy - dy*v.dx;
                 dx = c0;
        dy = c1;
        dz = c2;
    }
    
/**
  * Transforms "this" by a given transformation.
  * Sets the value of "this" to be "this" after applying the transformation "trans".
  * NB: Since "this" is a vector, translations have no effect.
  * DO NOT USE this method on
  * position vectors or surface-normal vectors.
  * @param trans The transformation
  */
    public void transform(Transformation trans) {
        double[][] m = trans.matrixd();
        double xnew, ynew, znew, wnew;
        xnew = m[0][0] * dx + m[0][1] * dy + m[0][2] * dz;
        ynew = m[1][0] * dx + m[1][1] * dy + m[1][2] * dz;
        znew = m[2][0] * dx + m[2][1] * dy + m[2][2] * dz;
        wnew = m[3][0] * dx + m[3][1] * dy + m[3][2] * dz;
        set(xnew, ynew, znew);
    }
}
  
